#import "FlutterServicePlugin.h"
#import "ATHContext.h"
#import "ATHMidService.h"
#import "ATHDeviceModel.h"
#import "HiidoSDK.h"
#import "NSArray+ATHFunctions.h"

// broadcast names

static NSString * const kServiceStatusBroadcast = @"flutter_service_broad_status";

static NSString * const kServiceKickOffBroadcast = @"flutter_service_broad_kickoff";

static NSString * const kServicePushDataBroadcast = @"flutter_service_broad_pushdata";

// method keys
@interface FlutterServicePlugin() <IYYMidServiceDelegate, FlutterStreamHandler>

@property (nonatomic, strong) id<IATHMidService> serviceChannel;
@property (nonatomic, copy) FlutterEventSink flutterEventCallback;

@end

@implementation FlutterServicePlugin
+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    FlutterMethodChannel* channel = [FlutterMethodChannel
                                     methodChannelWithName:@"flutter_service"
                                     binaryMessenger:[registrar messenger]];
    
    FlutterEventChannel *eventChannel = [FlutterEventChannel eventChannelWithName:@"flutter_service_event" binaryMessenger:[registrar messenger]];
    
    FlutterServicePlugin* instance = [[FlutterServicePlugin alloc] init];
    
    [eventChannel setStreamHandler:instance];
    [registrar addMethodCallDelegate:instance channel:channel];
}
    
- (void)handleMethodCall:(FlutterMethodCall*)call result:(FlutterResult)result {
    
    if (call.method.length == 0) {
        result(nil);
        return;
    }
    
    NSString *selName = [NSString stringWithFormat:@"%@:callback:", call.method];
    SEL sel = NSSelectorFromString(selName);
    if (sel && [self respondsToSelector: sel]) {
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Warc-performSelector-leaks"
        [self performSelector:sel withObject:call.arguments withObject:result];
#pragma clang diagnostic pop
        return ;
    }
    NSLog(@"Service Plugin cannot responds to %@", selName);
    result(nil);
}
    
- (void)initServiceWithParameters:(NSDictionary *)paramDict callback:(FlutterResult)callback
{
    if (self.serviceChannel) {
        callback(@YES);
        return;
    }
    NSLog(@"Init Service with params: %@", paramDict);
    id service = [ATHContext initMidwareForProtocol:@protocol(IATHMidService) withConfig:paramDict];
    self.serviceChannel = service;
    [self.serviceChannel addMidServiceDelegate:self];
    
    NSString *region = paramDict[@"region"];
    if ([region isKindOfClass: NSString.class] && region.length > 0) {
        [self.serviceChannel setCountryCode:region];
    }
    
    if (region == nil) {
        region = [ATHDeviceModel countryCode];
    }
    
    NSMutableDictionary *dict = @{}.mutableCopy;
    
    dict[@"country"] = region;
    dict[@"stype"] = @"2";
    dict[@"version"] = [ATHDeviceModel appVersion];
    dict[@"hdid"] = [HiidoSDK getHDID];
    dict[@"machine"] = [ATHDeviceModel modelName];
    dict[@"osVersion"] = [ATHDeviceModel systemVersion];
    dict[@"build"] = [ATHDeviceModel appBuild] ? : @"";
    
    [[ATH_MIDWARE(IATHMidService) getInstance] setDefaultHeaders:dict];
    [ATH_MIDWARE(IATHMidService) addClientHeaderEntries:dict];
    
    if (service) {
        callback(@YES);
    } else {
        callback(@NO);
    }
}

- (void)sendRequestWithParameters:(NSDictionary *)paramDict callback:(FlutterResult)callback
{
    YYServiceSendParams *params = [YYServiceSendParams new];
    FlutterStandardTypedData *typedData = paramDict[@"mRequestData"];
    
//    NSLog(@"Sending req: %@", paramDict);
    
    if (typedData && typedData.data) {
        params.server = paramDict[@"mServerName"];
        params.function = paramDict[@"mFunctionName"];
        params.prototype = paramDict[@"mProtoType"];
        params.headers = [paramDict[@"mClientHeaders"] isKindOfClass:NSDictionary.class] ? paramDict[@"mClientHeaders"] : @{};
        params.data = typedData.data;
    }
    
    [self.serviceChannel sendData:params completion:^(YYServiceReceivedData *rcvData) {
//        NSLog(@"Request completion with server [%@], func[%@], respData[%lu], svcerror: %@, pcserror: %@",rcvData.server, rcvData.function, rcvData.data.length, rcvData.serviceError, rcvData.processError);
        
        NSError *error = rcvData.serviceError ? : rcvData.processError;
        FlutterStandardTypedData *respData = nil;
        if (!error && rcvData.data) {
            respData = [FlutterStandardTypedData typedDataWithBytes:rcvData.data];
        }
        
        if (!respData) {
            error = error ? : [NSError errorWithDomain:@"FlutterServicePlugin Error" code:-1 userInfo:@{@"Reason": @"Unable to wrap nsdata to flutter typed data."}];
        }
        
        if (error || !respData) {
            callback(@{
                       @"error": error.description ? : @"UnknownError",
                       @"sdkResCode": @(rcvData.sdkResCode),
                       @"srvResCode": @(rcvData.sdkResCode),
                       @"reqId": @(rcvData.reqId)
                       });
            
        } else {
            callback(@{
                       @"mServerName": rcvData.server,
                       @"mFunctionName": rcvData.function,
                       @"mProtoType": rcvData.prototype,
                       @"mResponseData": respData,
                       @"mTraceId": rcvData.ORIGResponse.mTraceId ? : @"",
                       @"mResCode": rcvData.ORIGResponse.mResCode,
                       @"mServerHeaders": rcvData.ORIGResponse.mServerHeaders ? : @{},
                       @"mResMsg": rcvData.ORIGResponse.mResMsg ? : @"",
                       @"mBusinessContext": rcvData.ORIGResponse.mBusinessContext ? : @"",
                       });
        }
    }];
}

- (void)subscribeBroadcastWithParameters:(NSDictionary *)params callback:(FlutterResult)callback
{
    NSLog(@"Subscribing broad : %@", params);
    
    BroadScribeRequest *req = [self _broadRequestWithParameters:params];
    
    if (!req) {
        callback(@{
                   @"error": @"Invalid arguments",
                   @"sdkResCode": @-1,
                   @"srvResCode": @-1,
                   @"reqId": @0
                   });
    }
    
    
    [YYServiceCore.getInstance subScribeBroad:req option:nil sBlock:^(uint32_t reqId, BroadScribeResponse * _Nonnull response) {
        dispatch_async(dispatch_get_main_queue(), ^{
            callback(@{
                       @"reqId": @(reqId),
                       @"mBusinessContext": response.mBusinessContext?:@"",
                       @"mResCode": response.mResCode?:@0
                       });
        });
    } fBlock:^(uint32_t reqId, int sdkResCode, int srvResCode, NSError * _Nullable error) {
        dispatch_async(dispatch_get_main_queue(), ^{
            callback(@{
                       @"error": error.description ? : @"UnknownError",
                       @"sdkResCode": @(sdkResCode),
                       @"srvResCode": @(sdkResCode),
                       @"reqId": @(reqId)
                       });
        });
    }];

}


- (void)unsubscribeBroadcastWithParameters:(NSDictionary *)params callback:(FlutterResult)callback
{
    NSLog(@"UNSubscribing broad : %@", params);
    
    BroadScribeRequest *req = [self _broadRequestWithParameters:params];
    
    if (!req) {
        callback(@{
                   @"error": @"Invalid arguments",
                   @"sdkResCode": @-1,
                   @"srvResCode": @-1,
                   @"reqId": @0
                   });
    }
    
    [YYServiceCore.getInstance unSubScribeBroad:req option:nil sBlock:^(uint32_t reqId, BroadScribeResponse * _Nonnull response) {
        dispatch_async(dispatch_get_main_queue(), ^{
            callback(@{
                       @"reqId": @(reqId),
                       @"mBusinessContext": response.mBusinessContext?:@"",
                       @"mResCode": response.mResCode?:@0
                       });
        });
    } fBlock:^(uint32_t reqId, int sdkResCode, int srvResCode, NSError * _Nullable error) {
        dispatch_async(dispatch_get_main_queue(), ^{
            callback(@{
                       @"error": error.description ? : @"UnknownError",
                       @"sdkResCode": @(sdkResCode),
                       @"srvResCode": @(sdkResCode),
                       @"reqId": @(reqId)
                       });
        });
    }];
    
}




#pragma mark - Service Delegate

- (void)onMidServiceChannelStatus:(int)status
{
    [self broadcastToFlutterWithName:kServiceStatusBroadcast data:@(status)];
}

- (void)onMidServiceKickoff:(uint64_t)uid code:(int)code desc:(NSString *)desc
{
    [self broadcastToFlutterWithName:kServiceKickOffBroadcast
                                data:@{
                                       @"uid": @(uid),
                                       @"code": @(code),
                                       @"desc": desc ? : @""
                                       }];
}

- (void)onMidServiceReceivePushData:(YYServiceReceivedData *)receivedData
{
    NSLog(@"Did Rcv broad cast with grpType: %llu, id: %llu, data length: %llu", receivedData.groupType, receivedData.groupId, receivedData.data.length);
    
    NSData *d = receivedData.data ? : [NSData data];
    
    FlutterStandardTypedData *typedData = [FlutterStandardTypedData typedDataWithBytes:d];
    
    [self broadcastToFlutterWithName:kServicePushDataBroadcast
                                data:@{
                                       @"uid": @(receivedData.uid),
                                       @"groupType": @(receivedData.groupType),
                                       @"groupId": @(receivedData.groupId),
                                       @"server": receivedData.server ? : @"",
                                       @"function": receivedData.function ? : @"",
                                       @"prototype": receivedData.prototype ? : @"",
                                       @"data": typedData
                                       }];
}

#pragma mark - Flutter Stream Handler

- (FlutterError *)onListenWithArguments:(id)arguments eventSink:(FlutterEventSink)events
{
    if ([arguments isKindOfClass:NSString.class] && [@"flutter_service_event" isEqualToString:arguments]) {
        self.flutterEventCallback = events;
        return nil;
    }
    return [FlutterError errorWithCode:@"-1" message:[NSString stringWithFormat:@"Service Plugin unable to handle listen event for %@", arguments] details:nil];
}

- (FlutterError *)onCancelWithArguments:(id)arguments
{
    if ([arguments isKindOfClass:NSString.class] && [@"flutter_service_event" isEqualToString:arguments]) {
        self.flutterEventCallback = nil;
    }
    return nil;
}

#pragma mark - Helpers

- (void)broadcastToFlutterWithName:(NSString *)name data:(id)data
{
    if (name.length == 0 || !data || !self.flutterEventCallback) {
        return ;
    }
    
    NSLog(@"Broadcasting [%@]", name);
    self.flutterEventCallback(@{
                                @"broadcastName": name,
                                @"broadcastData": data
                                });
}

- (BroadScribeRequest *)_broadRequestWithParameters:(NSDictionary *)params
{
    
    NSString *businessContext = params[@"mBusinessContext"];
    NSArray <NSDictionary *> *rawUserGroups = params[@"mUserGrps"];
    
    if (rawUserGroups.count == 0) {
        return nil;
    }
    
    if ([businessContext isKindOfClass:NSNull.class]) {
        businessContext = @"";
    }
    
    NSArray *userGrps = [rawUserGroups ath_map:^id(NSDictionary *obj) {
        BroadUserGroup *grp = [BroadUserGroup new];
        grp.mGrpId = obj[@"mGrpId"];
        grp.mGrpType = obj[@"mGrpType"];
        return grp;
    }];
    
    BroadScribeRequest *req = [BroadScribeRequest new];
    req.mBusinessContext = businessContext;
    req.mUserGrps = userGrps.mutableCopy;
    
    return req;
}
    
@end

