import 'dart:async';
import 'dart:typed_data';

import 'package:flutter/services.dart';

const String kServiceStatusBroadcast = "flutter_service_broad_status";
const String kServiceKickOffBroadcast = "flutter_service_broad_kickoff";
const String kServicePushDataBroadcast = "flutter_service_broad_pushdata";

class FlutterServiceRequest {
  String businessContext;
  String serverName;
  String functionName;
  String protoType;
  Uint8List requestData;
  Map<String, String> routeArgs;
  Map<String, String> clientHeaders;
  String accountId;

  Map toJSON() {
    return {
      "mBusinessContext": businessContext,
      "mServerName": serverName,
      "mFunctionName": functionName,
      "mProtoType": protoType,
      "mRequestData": requestData,
      "mRouteArgs": routeArgs,
      "mClientHeaders": clientHeaders,
      "accountId":accountId,
    };
  }
}

class FlutterServiceResponse {
  String businessContext;
  num resCode;
  String resMsg;
  String serverName;
  String functionName;
  String protoType;
  Uint8List responseData;
  Map<String, dynamic> serverHeaders;
  String traceId;

  FlutterServiceResponse.fromJSON(Map jsonMap)
      : businessContext = jsonMap["mBusinessContext"],
        resCode = jsonMap["mResCode"],
        resMsg = jsonMap["mResMsg"],
        serverName = jsonMap["mServerName"],
        functionName = jsonMap["mFunctionName"],
        protoType = jsonMap["mProtoType"],
        responseData = new Uint8List.fromList(jsonMap["mResponseData"]) ,
        serverHeaders = new Map<String, dynamic>.from(jsonMap["mServerHeaders"]),
        traceId = jsonMap["mTraceId"];
}

class FlutterServicePushData {
  num uid;
  num groupType;
  num groupId;
  String serverName;
  String functionName;
  String protoType;
  Uint8List data;

  FlutterServicePushData.fromJSON(Map jsonMap)
      : uid = jsonMap["uid"],
        groupType = jsonMap["groupType"],
        groupId = jsonMap["groupId"],
        serverName = jsonMap["server"],
        functionName = jsonMap["function"],
        protoType = jsonMap['prototype'],
        data = new Uint8List.fromList(jsonMap["data"]);
}

class FlutterServiceUserGroup {
  num groupType;
  num groupId;

  Map toJSON() {
    return {"mGrpType": groupType, "mGrpId": groupId};
  }
}

class FlutterServiceBroadcastSubscribeRequest {
  String businessContext;
  List<FlutterServiceUserGroup> userGroups;

  Map toJSON() {
    List<Map> grps = userGroups.map((userGroup) {
      return userGroup.toJSON();
    }).toList();
    return {"mBusinessContext": businessContext, "mUserGrps": grps};
  }
}

class FlutterServiceBroadcastSubscribeResponse {
  String businessContext;
  num resCode;
  FlutterServiceBroadcastSubscribeResponse.fromJSON(
      Map jsonMap)
      : businessContext = jsonMap["mBusinessContext"],
        resCode = jsonMap["mResCode"];
}

class FlutterService {
  static const MethodChannel _channel = const MethodChannel('flutter_service');

  static const EventChannel _eventChannel =
      const EventChannel("flutter_service_event");

  static Function(String, Object) _broadcastCallback;
  static StreamSubscription _streamSubscription;

  static Future<bool> initService(Map map) async {
    bool success =
        await _channel.invokeMethod("initServiceWithParameters", map);
    return success;
  }

  static void _onRcvBroadcast(Object msg) {
    if (msg is Map && _broadcastCallback != null) {
      _broadcastCallback(msg["broadcastName"], msg["broadcastData"]);
    }
  }

  static void _onRcvBroadcastError(Object error) {
    print("On receive service error $error");
  }

  static void startReceivingBroadcasts(
      Function(String broadcastName, Object broadcastData) onReceiveBroadcast) {
    _broadcastCallback = onReceiveBroadcast;
    _streamSubscription = _eventChannel
        .receiveBroadcastStream("flutter_service_event")
        .listen(_onRcvBroadcast, onError: _onRcvBroadcastError);
  }

  static void stopReceivingBroadcasts() {
    if (_streamSubscription != null) {
      _streamSubscription.cancel();
      _streamSubscription = null;
      _broadcastCallback = null;
    }
  }

  static Future<void> request(
      FlutterServiceRequest request,
      Function(FlutterServiceResponse resp) onResp,
      Function(num reqId, num resCode, num serviceCode, String errorDesc) onFail
      ) async {
    var req = request.toJSON();
    Object res = await _channel.invokeMethod("sendRequestWithParameters", req);

    if (res is Map) {
      if (res["error"] != null) {
        onFail(res["reqId"], res["sdkResCode"], res["srvResCode"], res["error"]);
      } else {
        onResp(FlutterServiceResponse.fromJSON(res));
      }
      
    } else {
      onFail(0, -1, -1, "Invalid data type from native platform");
    }

    print("Request result : $res");
  }

  static Future<void> subscribeBroadcast(
    FlutterServiceBroadcastSubscribeRequest request,
    Function(num reqId, FlutterServiceBroadcastSubscribeResponse resp) onResp,
    Function(num reqId, num resCode, num serviceCode, String errorDesc) onFail
  ) async {
    Object res = await _channel.invokeMethod("subscribeBroadcastWithParameters", request.toJSON());
    if (res is Map) {
      if (res["error"] != null) {
        onFail(res["reqId"], res["sdkResCode"], res["srvResCode"], res["error"]);
      } else {
        onResp(res["reqId"], FlutterServiceBroadcastSubscribeResponse.fromJSON(res));
      }
    }
  }


  static Future<void> unsubscribeBroadcast(
    FlutterServiceBroadcastSubscribeRequest request,
    Function(num reqId, FlutterServiceBroadcastSubscribeResponse resp) onResp,
    Function(num reqId, num resCode, num serviceCode, String errorDesc) onFail
  ) async {
    Object res = await _channel.invokeMethod("unsubscribeBroadcastWithParameters", request.toJSON());
    if (res is Map) {
      if (res["error"] != null) {
        onFail(res["reqId"], res["sdkResCode"], res["srvResCode"], res["error"]);
      } else {
        onResp(res["reqId"], FlutterServiceBroadcastSubscribeResponse.fromJSON(res));
      }
    }
  }
}
